using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using PQ_SDK_MultiTouch;

using EPQT_Error = PQ_SDK_MultiTouch.PQMTClientImport.EnumPQErrorType;
using EPQT_TGesture = PQ_SDK_MultiTouch.PQMTClientImport.EnumPQTouchGestureType;
using EPQT_TPoint = PQ_SDK_MultiTouch.PQMTClientImport.EnumPQTouchPointType;
using EPQT_TRequest = PQ_SDK_MultiTouch.PQMTClientImport.EnumTouchClientRequestType;

namespace PQ_SDK_MultiTouchSample
{
    class Sample
    {
        delegate void PFuncOnTouchGesture(ref PQMTClientImport.TouchGesture tg);

        static PFuncOnTouchGesture[] g_pf_on_tges = new PFuncOnTouchGesture[(int)EPQT_TGesture.TG_TOUCH_END + 1];

        // use static delegate to prevent the GC collect the CallBack functors;
        static PQMTClientImport.PFuncOnReceivePointFrame cur_rf_func = new PQMTClientImport.PFuncOnReceivePointFrame(OnReceivePointFrame);
        static PQMTClientImport.PFuncOnReceiveGesture cur_rg_func = new PQMTClientImport.PFuncOnReceiveGesture(OnReceiveGesture);
        static PQMTClientImport.PFuncOnServerBreak cur_svr_break = new PQMTClientImport.PFuncOnServerBreak(OnServerBreak);
        static PQMTClientImport.PFuncOnReceiveError cur_rcv_err_func = new PQMTClientImport.PFuncOnReceiveError(OnReceiveError);
        static PQMTClientImport.PFuncOnGetServerResolution cur_get_resolution = new PQMTClientImport.PFuncOnGetServerResolution(OnGetServerResolution);
        static PQMTClientImport.PFuncOnGetDeviceInfo cur_fn_get_dev_info = new PQMTClientImport.PFuncOnGetDeviceInfo(OnGetDeviceInfo);

        static void Main(string[] args)
        {
            int err_code = Sample.InitAndConnectServer();
            if (err_code != (int)EPQT_Error.PQMTE_SUCCESS)
            {
                Console.WriteLine("press any key to exit...");
                Console.Read();
                return;
           }

            // do other things of your application;
            Console.WriteLine("hello world");
            // here, we just wait and not let the process exit;
            int read = 0;
            while (read != 'q' && read != 'Q')
            {
                Console.WriteLine("press \'q\' to exit");
                read = Console.Read();
            }
            Console.WriteLine("disconnect server...");
            PQMTClientImport.DisconnectServer();
            return;
        }
        static int InitAndConnectServer()
        {
            int err_code = (int)EPQT_Error.PQMTE_SUCCESS;
            try
            {
                // initialize
                Sample.InitFuncOnTG();

                // set the functions on receive
                Sample.SetFuncsOnReceiveProc();
                string local_ip = "127.0.0.1";

                // connect server
                Console.WriteLine(" connect to server...");
                if ((err_code = PQMTClientImport.ConnectServer(local_ip, PQMTClientImport.PQMT_DEFAULT_CLIENT_PORT)) != (int)EPQT_Error.PQMTE_SUCCESS)
                {
                    Console.WriteLine(" connect server fail, socket errror code:{0}", err_code);
                    return err_code;
                }
                // send request to server
                Console.WriteLine(" connect success, send request.");
                PQMTClientImport.TouchClientRequest tcq = new PQMTClientImport.TouchClientRequest();
                tcq.type = (int)EPQT_TRequest.RQST_RAWDATA_ALL |
                    (int)EPQT_TRequest.RQST_GESTURE_ALL;

                if ((err_code = PQMTClientImport.SendRequest(ref tcq)) != (int)EPQT_Error.PQMTE_SUCCESS)
                {
                    Console.WriteLine(" send request fail, error code:{0}",err_code);
                    return err_code;
                }
                ////////// you can set the move_threshold when the tcq.type is RQST_RAWDATA_INSIDE;
                //
                //int move_threshold = 1; // 1 pixel
                //err_code= PQMTClientImport.SendThreshold(move_threshold);
                //if (err_code != (int)EPQT_Error.PQMTE_SUCCESS)
                //{
                //    Console.WriteLine(" send threshold fail, error code:{0}",err_code);
                //}
                //
                //////// you can set the resolution of the touch point(raw data) here;
                //// setrawdata_resolution
                //int maxX = 32768, maxY = 32768;
                //if ((err_code = PQMTClientImport.SetRawDataResolution(maxX, maxY)) != (int)EPQT_Error.PQMTE_SUCCESS)
                //{
                //    Console.WriteLine(" set raw data resolution fail, error code:{0}", err_code);
                //}
                ////////////////////////
                //get server resolution
                if ((err_code = PQMTClientImport.GetServerResolution(OnGetServerResolution, IntPtr.Zero)) != (int)EPQT_Error.PQMTE_SUCCESS)
                {
                    Console.WriteLine(" get server resolution fail,error code:{0}" ,err_code);
                    return err_code;
                }
                // start receiving
                Console.WriteLine(" send request success, start recv.");

            }
            catch (Exception ex)
            {
                Console.WriteLine(" exception: {0}", ex.Message);
            }
            return err_code;
        }
        static void InitFuncOnTG()
        {
            // initialize the call back functions of toucha gestures;
            g_pf_on_tges[(int)EPQT_TGesture.TG_TOUCH_START] = (PFuncOnTouchGesture)OnTG_TouchStart;
            g_pf_on_tges[(int)EPQT_TGesture.TG_DOWN] = (PFuncOnTouchGesture)OnTG_Down;
            g_pf_on_tges[(int)EPQT_TGesture.TG_MOVE] = (PFuncOnTouchGesture)OnTG_Move;
            g_pf_on_tges[(int)EPQT_TGesture.TG_UP] = (PFuncOnTouchGesture)OnTG_Up;

            g_pf_on_tges[(int)EPQT_TGesture.TG_SECOND_DOWN] = (PFuncOnTouchGesture)OnTG_SecondDown;
            g_pf_on_tges[(int)EPQT_TGesture.TG_SECOND_UP] = (PFuncOnTouchGesture)OnTG_SecondUp;

            g_pf_on_tges[(int)EPQT_TGesture.TG_SPLIT_START] = (PFuncOnTouchGesture)OnTG_SplitStart;
            g_pf_on_tges[(int)EPQT_TGesture.TG_SPLIT_APART] =(PFuncOnTouchGesture) OnTG_SplitApart;
            g_pf_on_tges[(int)EPQT_TGesture.TG_SPLIT_CLOSE] = (PFuncOnTouchGesture)OnTG_SplitClose;
            g_pf_on_tges[(int)EPQT_TGesture.TG_SPLIT_END] = (PFuncOnTouchGesture)OnTG_SplitEnd;

            g_pf_on_tges[(int)EPQT_TGesture.TG_TOUCH_END] = (PFuncOnTouchGesture)OnTG_TouchEnd;
        }

        static void SetFuncsOnReceiveProc()
        {
            PQMTClientImport.SetOnReceivePointFrame(cur_rf_func,IntPtr.Zero);
            PQMTClientImport.SetOnReceiveGesture(cur_rg_func, IntPtr.Zero);
            PQMTClientImport.SetOnServerBreak(cur_svr_break, IntPtr.Zero);
            PQMTClientImport.SetOnReceiveError(cur_rcv_err_func, IntPtr.Zero);
            PQMTClientImport.SetOnGetDeviceInfo(cur_fn_get_dev_info, IntPtr.Zero);
        }

        static void OnReceivePointFrame(int frame_id, int time_stamp, int moving_point_count, IntPtr moving_point_array,IntPtr call_back_object)
        {
            string[] tp_event = new string[3]
            {
                "down",
                "move",
                "up"
            };
            Console.WriteLine("frame_id:{0},time_stamp:{1} ms,moving point count:{2}", frame_id, time_stamp, moving_point_count);
            for (int i = 0; i < moving_point_count; ++i)
            {
                IntPtr p_tp = (IntPtr)(moving_point_array.ToInt64() + i * Marshal.SizeOf(typeof(PQMTClientImport.TouchPoint)));
                PQMTClientImport.TouchPoint tp = (PQMTClientImport.TouchPoint)Marshal.PtrToStructure(p_tp, typeof(PQMTClientImport.TouchPoint));
                
			    OnTouchPoint(ref tp);
	        }
        }
        static void OnReceiveGesture(ref PQMTClientImport.TouchGesture gesture, IntPtr call_back_object)
        {
            OnTouchGesture(ref gesture);
        }

        static void OnServerBreak(IntPtr param,IntPtr call_back_object)
        {
	        // when the server break, disconenct server;
            Console.WriteLine("server break, disconnect here");
            PQMTClientImport.DisconnectServer();
        }
        static void OnGetServerResolution(int x, int y, IntPtr call_back_object)
        {
            Console.WriteLine("server resolution:{0},{1}", x, y);
        }
        static void OnReceiveError(int err_code,IntPtr call_back_object)
        {
            switch (err_code)
            {
                case (int)EPQT_Error.PQMTE_RCV_INVALIDATE_DATA:
                    Console.WriteLine(" error: receive invalidate data.");
                    break;
                case (int)EPQT_Error.PQMTE_SERVER_VERSION_OLD:
                    Console.WriteLine(" error: the multi-touch server is old for this client, please update the multi-touch server.");
                    break;
                case (int)EPQT_Error.PQMTE_EXCEPTION_FROM_CALLBACKFUNCTION:
                    Console.WriteLine(" **** some exceptions thrown from the call back functions.");
                    break;
                default:
                    Console.WriteLine(" socket error, socket error code:{0}", err_code);
                    break;
            }
        }
        static void OnGetDeviceInfo(ref PQMTClientImport.TouchDeviceInfo deviceInfo,IntPtr call_back_object){
	        Console.WriteLine(" touch screen, SerialNumber:{0},({1},{2}). ",deviceInfo.serial_number,deviceInfo.screen_width,deviceInfo.screen_height);
        }
        // here, just record the position of point,
        //	you can do mouse map like "OnTG_Down" etc;
        static void OnTouchPoint(ref PQMTClientImport.TouchPoint tp)
        {
            switch ((EPQT_TPoint)tp.point_event)
	        {
	        case EPQT_TPoint.TP_DOWN:
                Console.WriteLine("  point {0} come at ({1},{2}) width:{3} height:{4}", tp.id, tp.x, tp.y, tp.dx, tp.dy);
		        break;
	        case EPQT_TPoint.TP_MOVE:
                Console.WriteLine("  point {0} move at ({1},{2}) width:{3} height:{4}", tp.id, tp.x, tp.y, tp.dx, tp.dy);
		        break;
	        case EPQT_TPoint.TP_UP:
                Console.WriteLine("  point {0} leave at ({1},{2}) width:{3} height:{4}", tp.id, tp.x, tp.y, tp.dx, tp.dy);
		        break;
	        }
        }

        static void OnTouchGesture(ref PQMTClientImport.TouchGesture tg)
        {
	        if((int)EPQT_TGesture.TG_NO_ACTION == tg.type)
		        return ;
        	
	        DefaultOnTG(ref tg);

	        PFuncOnTouchGesture pf = (PFuncOnTouchGesture) g_pf_on_tges[(int)tg.type];
	        if(null != pf)
	        {
		        pf(ref tg);
	        }
        }


        static void OnTG_TouchStart(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_TOUCH_START);
            Console.WriteLine("  here, the touch start, initialize something.");
        }

        static void DefaultOnTG(ref PQMTClientImport.TouchGesture tg) // just show the gesture
        {
            string name = Marshal.PtrToStringAnsi(PQMTClientImport.GetGestureName(ref tg));
            Console.Write("gesture:{0},type:{1},param size:{2} ", name,tg.type, tg.param_size);
            for (int i = 0; i < tg.param_size; ++i)
            {
                //double tmp = 0;
                //tg.param_s.GetValue(i,out tmp);
                Console.Write("{0} ", tg.param_s[i]);
            }
	        Console.WriteLine("");
        }

        static void OnTG_Down(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_DOWN && tg.param_size >= 2);

	        Console.WriteLine("  the single finger touching at :({0},{1})", 
		        tg.param_s[0],tg.param_s[1]);
        }

        static void OnTG_Move(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_MOVE && tg.param_size >= 2);

	        Console.WriteLine("  the single finger moving on the screen at :({0},{1}) ",
                tg.param_s[0], tg.param_s[1]);
        }
        static void OnTG_Up(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_UP && tg.param_size >= 2);

	        Console.WriteLine(" the single finger is leaving the screen at :({0},{1})",
		        tg.param_s[0] ,tg.param_s[1]);
        }

        //
        static void OnTG_SecondDown(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_SECOND_DOWN && tg.param_size >= 4);

            Console.WriteLine("  the second finger touching at :({0},{1}),  after the first finger touched at :({2},{3})",
                tg.param_s[0], tg.param_s[1], tg.param_s[2], tg.param_s[3]);
        }

        static void OnTG_SecondUp(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_SECOND_UP && tg.param_size >= 4);

            Console.WriteLine("  the second finger is leaving at :({0},{1}) while the first finger still anchored around :({2},{3}) ",
                tg.param_s[0], tg.param_s[1],
                tg.param_s[2], tg.param_s[3]);
        }

        //
        static void OnTG_SplitStart(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_SPLIT_START && tg.param_size >= 4);

	        Console.WriteLine("  the two fingers is splitting with one finger at: ({0},{1}),the other at :({2},{3})",
		        tg.param_s[0] ,tg.param_s[1],
		        tg.param_s[2] ,tg.param_s[3]);
        }

        static void OnTG_SplitApart(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_SPLIT_APART && tg.param_size >= 1);

	        Console.WriteLine("  the two fingers is splitting apart with there distance incresed by {0} with a ratio :{1} ", 
		        tg.param_s[0],tg.param_s[1]);
        }
        static void OnTG_SplitClose(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_SPLIT_CLOSE && tg.param_size >= 1);

	        Console.WriteLine("  the two fingers is splitting close with there distance decresed by {0} with a ratio:{1}",
		        tg.param_s[0],tg.param_s[1]);
        }

        static void OnTG_SplitEnd(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_SPLIT_END);

            Console.WriteLine("  the two splitting fingers with one finger at: ({0},{1}),the other at :({2},{3} will end",
                tg.param_s[0], tg.param_s[1],
                tg.param_s[2], tg.param_s[3]);
        }

        // OnTG_TouchEnd: to clear what need to clear
        static void OnTG_TouchEnd(ref PQMTClientImport.TouchGesture tg)
        {
            Debug.Assert(tg.type == (int)EPQT_TGesture.TG_TOUCH_END);

            Console.WriteLine("  all the fingers is leaving and there is no fingers on the screen.");
        }
    }
}
